
/*
   SBDSP - a program to allow the use of SoundBlaster audio I/O boards
	   for APT satellite image capture/demodulation.

	   Written by:	Mark S. Sims
			Box 1601 SMU
			Dallas, TX 75275

	   Compiled with: Microsoft Quick C 2.0 / MSC 6.0

	   Copyright (C) 1992 by Mark S. Sims.
	   This program may be freely copied and distributed as long
	   as this message stays intact.  Commercial or for-profit
	   distribution,  modification,  or use is strictly prohibited.

	   Use and modify at your own risk 'til your heart's content
	   as long as all modifications remain in the public domain.

	   A good place to start would be to add automatic image centering
	   or a software PLL to remove the dopple bow.	Or perhaps an
	   optional noise filter.  Or calculate the demodulator table
	   with integer math and can the floating point library.  Or ...
	   well you get the point...


      Version 1.0 - Initial release
*/

#define VERSION 0x10

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <io.h>
#include <dos.h>
#include <malloc.h>
#include <stdlib.h>
#include <conio.h>
#include <math.h>

#pragma pack(1)    /* we need packed structures! */

#define ABS(x) ((x<0)? (0-x): x)
#define SIGN(x) ((x<0)? '-': '+')

int	dma_odd = 2;		 /* add two pixels to ODD DMA input blocks */
int	dma_even = 3;		 /* and three pixels to the EVEN blocks */
int	coarse_tune = (-4808/8); /* and remove every 601st pixel from the stream */
long	fine_tune = (0);	 /* +40000 would duplicate every 40,000th pixel */
				 /* -12345 would remove every 12,345th pixel, etc */

unsigned char in_name[128];	 /* input file name */
FILE *finfile;			 /* input file handle */
long readpos;			 /* input file read position */

unsigned char out_name[128];	 /* output file name */
FILE *foutfile; 		 /* output file handle */
long writepos;			 /* output file write position */

struct HEADER {       /* SoundBlaster .VOC file DMA block header */
   unsigned char id;
   unsigned short block_size;
   unsigned char size_high;
   unsigned short sample_rate;
} header;

#define MAX_BLOCK 32768       /* maximum DMA input block size */
#define MAX_DMA_ADJUST 256    /* how much adjustment we will allow */
unsigned blocks;	      /* the input block counter */

unsigned char in_buf[MAX_BLOCK + MAX_DMA_ADJUST + 2];
unsigned read_cols = (MAX_BLOCK - sizeof(struct HEADER));

int calc_average = 1;	/* set flag to calculate the block sample averages */
long average;		/* where we calculate the average */
unsigned char bias;	/* the resulting DC offset of the SoundBlaster samples */

#define BUFLEN	    (512*8)	 /* DOS writes multiples of 512 bytes fastest */
unsigned char out_buf[BUFLEN];	 /* output file buffer */
unsigned outptr;		 /* output buffer pointer */

unsigned char huge *demod;    /* the demodulator table is allocated here */
unsigned long demod_size;     /* it is this many bytes long */



main(argc, argv)
int  argc;
char *argv[];
{
int i;
int key;

   key = ' ';
   for(i=1; i<argc; i++) {  /* process the command line file names */
      if((argv[i][0] != '/') && (argv[i][0] != '-')) set_file(argv[i]);
   }
   if(in_name[0] == 0) goto help;   /* no input file was given */

   for(i=1; i<argc; i++) {  /* process the command options */
      if((argv[i][0] == '/') || (argv[i][0] == '-')) {
	 key = option_switch(argv[i]);
	 if(key) {
	    help:
	    printf("SBDSP - a program to demodulate APT satellite images from\n");
	    printf("        SoundBlaster .VOC audio sample files.  Version %X.%X\n", VERSION>>4, VERSION&0x0F);
	    printf("\n");
	    printf("        Copyright (C) 1992 by Mark S. Sims.\n");
	    printf("        This program may be freely copied and distributed as long\n");
	    printf("        as this message stays intact.  Commercial or for-profit\n");
	    printf("        distribution or use is strictly prohibited.\n");
	    printf("\n");
	    printf("*** ERROR: Invalid command line option: %c.  Usage is:\n", key);
	    printf("      sbdsp infile [outfile] [/E+val] [/O-val] [/C+val] [/F-val]\n");
	    printf("            /E - set even DMA block adjustment factor\n");
	    printf("                 (default: /E%c%d)\n", ((dma_even<0) ? '-': '+'), ABS(dma_even));
	    printf("            /O - set odd DMA block adjustment factor\n");
	    printf("                 (default: /O%c%d)\n", ((dma_odd<0) ? '-': '+'), ABS(dma_odd));
	    printf("            /C - set coarse (line) timing adjustment factor\n");
	    printf("                 (default: /C%c%d)\n", ((coarse_tune<0) ? '-': '+'), ABS(coarse_tune));
	    printf("            /F - set fine (image skew) timing adjustment factor\n");
	    printf("                 (default: /F%c%ld)\n", ((fine_tune<0) ? '-': '+'), ABS(fine_tune));
	    printf("\n");
	    printf("            If *outfile* is not given, the demodulated output image\n");
	    printf("            is written over the input sample file (VERY SLOWLY)!\n");
	    exit(1);
	 }
      }
   }

   setup_files();
   build_demodulator();
   demodulate_file();

   printf("\n\nDONE!!!  Demodulated %d blocks\n", blocks);
   fclose(finfile);
   if(finfile != foutfile) fclose(foutfile);
   exit(0);
}


set_file(arg)
char arg[];
{
   if(in_name[0]) strcpy(out_name, arg);
   else strcpy(in_name, arg);

   return 0;
}


option_switch(arg)
char arg[];
{
char c;
int s;
int i;

   i = 1;
   c = tolower(arg[i]);     /* the option character */
   s = arg[++i];	    /* the sign (+ or -) */
   if(s == '=') {
      s = arg[++i];
   }
   if((s != '+') && (s != '-')) s = 0;

   if(c == 'e') {
      if(s == 0) return c;
      sscanf(&arg[i], "%d", &dma_even);
      if(ABS(dma_even) > MAX_DMA_ADJUST) return c;
   }
   else if(c == 'o') {
      if(s == 0) return c;
      sscanf(&arg[i], "%d", &dma_odd);
      if(ABS(dma_odd) > MAX_DMA_ADJUST) return c;
   }
   else if(c == 'c') {
      if(s == 0) return c;
      sscanf(&arg[i], "%d", &coarse_tune);
   }
   else if(c == 'f') {
      if(s == 0) return c;
      sscanf(&arg[i], "%ld", &fine_tune);
   }
   else return c;

   return 0;
}


setup_files()
{
unsigned j;
char c;

   finfile = fopen(in_name, "r+b");
   if(finfile == 0) {
      printf("*** ERROR: could not open input file for fread(): %s\n", in_name);
      exit(1);
   }
   j = fread(in_buf, 1, 26, finfile);
   if(j != 26) {
      printf("*** ERROR: could not skip header on input file: %s\n", in_name);
      exit(1);
   }

   if(out_name[0]) {	/* output file name was given, write a separate file */
      foutfile = fopen(out_name, "wb");
      if(foutfile == 0) {
	 printf("*** ERROR: could not open output file for fwrite(): %s\n", out_name);
	 exit(1);
      }
   }
   else {   /* no output file name was given,  overwrite the input */
      while(1) {
	 printf("WARNING: No output file name was given.  Overwrite input (Y/N)? ");
	 c = getch();
	 c = tolower(c);
	 printf("\n");

	 if(c == 'n') {
	    printf("OK. I'wont...\n");
	    exit(1);
	 }
	 else if(c == 'y') {
	    printf("Overwriting the input file.  Interruption is not recommeded...\n");
	    break;
	 }
	 else printf("\n");
      }

      foutfile = finfile;
      readpos = ftell(finfile);  /* where we need to start reading the sample data */
   }

   return 0;
}


build_demodulator()
{
long x, y;
double fscale;
unsigned long index;	/* used to access the demodulator table */
unsigned xm;

   demod_size = 256L*256L;
   demod = halloc(demod_size, 1);   /* this is a 64Kb array! */
   if(demod == 0) {
      printf("*** ERROR: could not halloc() %ld bytes of memory.\n", demod_size);
      exit(1);
   }

   fscale = 255.0 / sqrt(128.0*128.0 + 128.0*128.0);
   printf("Initializing demodulator tables...");
   for(x=0; x<256; x++) {
      for(y=0; y<256; y++) {
	 index = ((x-128) * (x-128)) + ((y-128) * (y-128));
	 xm = (unsigned) (sqrt((double) index) * fscale);
	 if(xm > 255) xm = 255;

	 index = (x << 8) | y;
	 if(index >= demod_size) {
	    printf("\n*** ERROR: demoulator table overflow.\n");
	    exit(1);
	 }
	 demod[index] = xm;
      }
   }

   printf("\n");
   return 0;
}


demodulate_file()
{
unsigned i, j;
int coarse_time;    /* coarse skip rate */
long fine_time;     /* fine skip rate */
int dma_add;	    /* end-of-DMA block pixel compensator */
unsigned short index;

   printf("\n");
   if(finfile == foutfile) printf("DON'T press ESC to interrupt...\n");
   else printf("Press ESC to interrupt...\n");
   printf("\n");

   printf("Demodulating %d byte blocks...\n", read_cols);
   printf(
      "   /O%c%d  /E%c%d  /C%c%d  /F%c%ld\n",
      SIGN(dma_odd),	 ABS(dma_odd),
      SIGN(dma_even),	 ABS(dma_even),
      SIGN(coarse_tune), ABS(coarse_tune),
      SIGN(fine_tune),	 ABS(fine_tune)
   );

   blocks = 0;
   outptr = 0;
   coarse_time = 0;
   fine_time = 0;

   while(1) {
      if(kbhit()) {   /* the user pressed a key */
	 if(getch() == 0x1B) {	 /* was it ESC? */
	    printf("\n*** INTERRUPTED ***\n");
	    break;
	 }
	 else printf("\007");  /* beep at the fumble fingered klutz */
      }

      if(finfile == foutfile) fseek(finfile, readpos, 0);
      j = fread(&header, 1, sizeof(struct HEADER), finfile);  /* block header */
      if(j != sizeof(struct HEADER)) {
	 break;
      }
      if(header.id != 0x01) {
	 printf("Block header ID %02X is not supported.\n", header.id);
	 break;
      }
      else if(header.size_high || (header.block_size > MAX_BLOCK)) {
	 printf("Input file block size too big.\n");
	 break;
      }

      /* the first word in the block (which we actually read in as part */
      /* of the block header) seems to be the sample clock rate code. */
      /* We need to remove it from the block size value... */
      read_cols = header.block_size - 2;

      j = fread(in_buf, 1, read_cols, finfile);  /* get data buffer values */
      if(j != read_cols) break;    /* its close to the end-of-file */
      readpos = ftell(finfile);

      /* duplicate the last few samples in the block to compensate for those */
      /* lost due to the time VREC spends setting up between DMA blocks */
      if(blocks & 0x01) dma_add = dma_odd; /* ODD blocks need this many */
      else dma_add = dma_even;	      /* and EVEN blocks need this many */

      if(dma_add > 0) {   /* we are adding pixels to this DMA block */
	 for(i=0; i<dma_add; i++) {  /* duplicate the last few pixels */
	    in_buf[read_cols+i] = in_buf[read_cols-dma_add+i];
	 }
      }
      read_cols += dma_add;  /* adjust the resulting pixel count */

      /* calculate the average sample value in the block */
      if(calc_average) {
	 /* calc_average = 0; */  /* uncomment if your board's sample bias does not drift */
	 average = 0;
	 for(i=0; i<read_cols; i++) average += in_buf[i];
	 average += (read_cols / 2);   /* round it */
	 average /= read_cols;
	 bias = (average - 128L);   /* the samples have a net bias */
      }

      /* demodulate the block of samples to convert the input APT sine */
      /* wave sample into their corresponding pixel values */
      for(i=0; i<read_cols-1; i++) {
	 if((coarse_tune < 0) && (--coarse_time <= coarse_tune)) {
	    coarse_time = 0;   /* we are removing excess samples from the line */
	 }
	 else if((fine_tune < 0) && (--fine_time <= fine_tune)) {
	    fine_time = 0;   /* we are removing pixels to fine tune the clock rate */
	 }
	 else {  /* demodulate the remaining samples */
	    index = ((unsigned short) (in_buf[i] - bias));
	    index <<= 8;
	    index |= ((unsigned short) (in_buf[i+1] - bias));

	    out_buf[outptr] = demod[index];

	    if(++outptr >= BUFLEN) {   /* its time to write the output buffer */
	       if(finfile == foutfile) {
		  fseek(foutfile, writepos, 0);
		  j = fwrite(out_buf, 1, outptr, foutfile);
		  writepos = ftell(foutfile);
	       }
	       else {
		  j = fwrite(out_buf, 1, outptr, foutfile);
	       }
	       if(j != outptr) return 1;   /* !!! file write error */
	       outptr = 0;
	    }

	    /* see if we are duplicating pixels */
	    if((coarse_tune > 0) && (++coarse_time >= coarse_tune)) {
	       coarse_time = 0;   /* adjusting the raw sample rate */
	       --i;		  /* by duplicating the last pixel */
	    }
	    else if((fine_tune > 0) && (++fine_time >= fine_tune)) {
	       fine_time = 0;	  /* fine tuning the clock */
	       --i;		  /* by duplicating the last pixel */
	    }
	 }
      }

      ++blocks;
      printf("  demodulated %d blocks (%ld bytes) so far.  bias=%d.\r", blocks, readpos, bias);
   }

   if(outptr) {   /* write any data remaining in the output buffer */
      if(finfile == foutfile) {   /* overwriting the input file with output data */
	 fseek(foutfile, writepos, 0);
	 j = fwrite(out_buf, 1, outptr, foutfile);
	 writepos = ftell(foutfile);
      }
      else {
	 j = fwrite(out_buf, 1, outptr, foutfile);
      }
   }

   return 0;
}

